const std = @import("std");
const toks = @import("tokenizer.zig");

pub const ParseError = error{ no_fn_name, no_fn_params };

pub const Tag = enum {
    ///expression
    ///`<rhs...>`
    ///
    ///lhs ignored
    expr,

    ///variable assignment
    ///`<lhs> = <rhs...>`
    ///
    ///lhs is overwritten to be variable
    var_assign,

    ///lazy evaluated list
    ///`'(<rhs...>)`
    ///
    ///lhs ignored
    llist,

    ///list (evaluated)
    ///`(<lhs> <rhs...>)`
    ///
    ///lhs needs to be a callable
    list_eval,

    ///fn_def (procedure)
    ///`;;; <lhs>
    ///(procedure <main_token>(<lhs>) <rhs...>)`
    fn_def,
};

pub const Node = struct {
    tag: Tag,
    main_token: Index,
    data: Data,

    pub const Data = struct {
        lhs: Index,
        rhs: Index,
    };
    pub const Index = u32;
};
pub const AstError = error{};

pub const Parser = struct {
    gpa: std.mem.Allocator,
    source: [:0]const u8,

    token_tags: []const toks.Token.Tag,
    token_locs: []const toks.Token.Loc,
    tok_i: Node.Index,

    errs: std.ArrayList(AstError),
    nodes: std.MultiArrayList(Node),
    extra_data: std.ArrayList(Node.Index),
    scratch: std.ArrayList(Node.Index),

    pub fn init(buffer: [:0]const u8, mal: std.MultiArrayList(toks.Token), allocator: std.mem.Allocator) !Parser {
        return .{
            .gpa = allocator,
            .source = buffer,

            .token_tags = mal.items(.tag),
            .token_locs = mal.items(.loc),
            .tok_i = 0,

            .errs = std.ArrayList(AstError).init(allocator),
            .nodes = std.MultiArrayList(Node){},
            .extra_data = std.ArrayList(Node.Index).init(allocator),
            .scratch = std.ArrayList(Node.Index).init(allocator),
        };
    }

    fn hasToken(self: *Parser, expected: toks.Token.Tag, offset: isize) ?toks.Token {
        if (self.token_tags[@intCast(self.tok_i + offset)] == expected) {
            return .{ .loc = self.token_locs[@intCast(self.tok_i + offset)], .tag = self.token_tags[@intCast(self.tok_i + offset)] };
        }
        return null;
    }
    fn eatToken(self: *Parser, expected: toks.Token.Tag) ?Node.Index {
        const tok = self.hasToken(expected, 0);
        if (tok != null) {
            self.tok_i += 1;
            return self.tok_i - 1;
        }
        return null;
    }

    fn parse_fn_proc(self: *Parser) ?Node {
        _ = self.eatToken(.sym);
        if (self.hasToken(.list_l, -2) != null) {
            // lisp style
        } else if (self.eatToken(.list_l) != null) {
            // c style
        } else {
            // not a procedure call or invalid syntax?
        }

        const name = self.eatToken(.sym) orelse return null;
        std.debug.print("found procedure def for `{s}`", .{self.source[self.token_locs[name].start..self.token_locs[name].end]});
        _ = self.eatToken(.list_l) orelse return null;
        var open_lists: usize = 0;
        while (true) : (self.tok_i += 1) {
            switch (self.token_tags[self.tok_i]) {
                .list_l, .list_lz => {
                    open_lists += 1;
                },
                .list_r => {
                    if (open_lists > 0) {
                        open_lists -= 1;
                    } else {
                        break;
                    }
                },
                else => {},
            }
        }

        while (true) : (self.tok_i += 1) {
            switch (self.token_tags[self.tok_i]) {
                .list_l, .list_lz => {
                    open_lists += 1;
                },
                .list_r => {
                    if (open_lists > 0) {
                        open_lists -= 1;
                    } else {
                        break;
                    }
                },
                else => {},
            }
        }
        self.tok_i += 1;

        return Node{ .tag = .fn_def, .main_token = name, .data = .{ .lhs = 0, .rhs = 0 } };
    }

    pub fn next(self: *Parser) ?Node {
        while (self.tok_i < self.token_tags.len) : (self.tok_i += 1) {
            switch (self.token_tags[self.tok_i]) {
                toks.Token.Tag.sym => {
                    if (std.mem.eql(u8, "procedure", self.source[self.token_locs[self.tok_i].start..self.token_locs[self.tok_i].end])) {
                        return self.parse_fn_proc();
                    }
                },
                else => {},
            }
        }
        return null;
    }
};
test "parsing of simple example" {
    const example =
        \\t
        \\nil
        \\a = b
        \\"some string w/ escaped\""
        \\(procedure a() )
    ;

    var tokz = toks.Tokenizer.init(example);
    var tokens = std.MultiArrayList(toks.Token){};
    defer tokens.deinit(std.testing.allocator);
    while (tokz.next()) |tok| {
        try tokens.append(std.testing.allocator, tok);
        std.debug.print("{}\n", .{tok});
    }
    var parse = try Parser.init(example, tokens, std.testing.allocator);
    while (parse.next()) |ast_node| {
        std.debug.print("{}\n", .{ast_node});
    }
}
