const std = @import("std");

pub const Token = struct {
    tag: Tag,
    loc: Loc,

    pub const Loc = struct {
        start: usize,
        end: usize,
    };

    pub const Tag = enum {
        sym,
        num,
        str,
        /// t
        t,
        /// nil
        nil,
        /// =
        assign,
        /// -=
        assign_sub,
        /// /=
        assign_div,
        /// *=
        assign_mul,
        /// +=
        assign_add,
        /// ==
        op_eq,
        /// >
        op_gt,
        /// >=
        op_geq,
        /// <
        op_lt,
        /// <=
        op_leq,
        /// /
        op_div,
        /// *
        op_mul,
        /// +
        op_add,
        /// -
        op_sub,
        /// ->
        op_acc,
        /// ~>
        op_derefacc,
        /// %
        op_mod,
        /// !
        op_not,
        /// !=
        op_neq,
        /// ||
        op_or,
        /// &&
        op_and,
        /// (
        list_l,
        /// '(
        list_lz,
        /// )
        list_r,
        /// @keys
        kw_keys,
        /// @rest
        kw_rest,
    };

    pub fn format(self: *const Token, comptime _: []const u8, _: std.fmt.FormatOptions, writer: anytype) !void {
        try writer.print("{d}:{d} .{s}", .{ self.loc.start, self.loc.end, @tagName(self.tag) });
    }
};

pub const Tokenizer = struct {
    buffer: [:0]const u8,
    index: usize,
    start: usize,

    const State = enum {
        start,
        alphanum_identifier,
        number_or_float,
        decimals,
        signed_exponent,
        unsigned_exponent,
        string,
        op_plus,
        op_minus,
        op_star,
        op_fslash,
        op_pipe,
        op_amp,
        op_excl,
        op_deref,
        op_eq,
        list_l,
        list_lz,
        list_r,
    };

    pub fn init(buf: [:0]const u8) Tokenizer {
        return .{
            .buffer = buf,
            .index = 0,
            .start = 0,
        };
    }

    pub fn next(self: *Tokenizer) ?Token {
        var state: State = .start;
        while (self.index < self.buffer.len) : (self.index += 1) {
            const c = self.buffer[self.index];
            const loc = Token.Loc{ .start = self.start, .end = self.index };
            state = switch (state) {
                .start => blk: {
                    self.start = self.index;
                    break :blk switch (c) {
                        'a'...'z', 'A'...'Z', '_' => .alphanum_identifier,
                        '0'...'9' => .number_or_float,
                        '.' => .decimals,
                        '"' => .string,
                        '+' => .op_plus,
                        '-' => .op_minus,
                        '*' => .op_star,
                        '/' => .op_fslash,
                        '|' => .op_pipe,
                        '&' => .op_amp,
                        '!' => .op_excl,
                        '~' => .op_deref,
                        '=' => .op_eq,
                        '(' => .list_l,
                        ')' => .list_r,
                        '\'' => .list_lz,
                        else => .start,
                    };
                },
                .alphanum_identifier => switch (c) {
                    'a'...'z', 'A'...'Z', '0'...'9', '_' => .alphanum_identifier,
                    else => {
                        inline for (.{ Token.Tag.t, Token.Tag.nil }) |alphanum_tag| {
                            if (std.mem.eql(u8, self.buffer[self.start..self.index], @tagName(alphanum_tag))) {
                                return Token{ .tag = alphanum_tag, .loc = loc };
                            }
                        }
                        return Token{ .tag = .sym, .loc = loc };
                    },
                },
                .number_or_float => switch (c) {
                    '0'...'9' => .number_or_float,
                    '.' => .decimals,
                    'e' => .signed_exponent,
                    ' ', '\n' => {
                        return Token{ .tag = .num, .loc = loc };
                    },
                    else => unreachable,
                },
                .decimals => switch (c) {
                    '0'...'9' => .decimals,
                    ' ', '\n' => {
                        return Token{ .tag = .num, .loc = loc };
                    },
                    else => unreachable,
                },
                .signed_exponent => switch (c) {
                    '0'...'9', '+', '-' => .unsigned_exponent,
                    else => unreachable,
                },
                .unsigned_exponent => switch (c) {
                    '0'...'9' => .unsigned_exponent,
                    ' ', '\n' => {
                        return Token{ .tag = .num, .loc = loc };
                    },
                    else => unreachable,
                },
                .string => switch (c) {
                    '"' => {
                        return Token{ .tag = .str, .loc = loc };
                    },
                    '\\' => blk: {
                        self.index += 1;
                        break :blk .string;
                    },
                    else => .string,
                },
                .op_plus, .op_minus, .op_fslash, .op_star, .op_excl, .op_eq => switch (c) {
                    '=' => {
                        return Token{ .tag = switch (state) {
                            .op_plus => .assign_add,
                            .op_minus => .assign_sub,
                            .op_star => .assign_mul,
                            .op_fslash => .assign_div,
                            .op_excl => .op_neq,
                            .op_eq => .op_eq,
                            else => unreachable,
                        }, .loc = loc };
                    },
                    ' ', '\n' => {
                        return Token{ .tag = switch (state) {
                            .op_plus => .op_add,
                            .op_minus => .op_sub,
                            .op_star => .op_mul,
                            .op_fslash => .op_div,
                            .op_excl => .op_not,
                            .op_eq => .assign,
                            else => unreachable,
                        }, .loc = loc };
                    },
                    '>' => {
                        return Token{ .tag = switch (state) {
                            .op_minus => .op_acc,
                            else => unreachable,
                        }, .loc = loc };
                    },
                    else => unreachable,
                },
                .op_pipe => switch (c) {
                    '|' => {
                        return Token{ .tag = .op_or, .loc = loc };
                    },
                    else => unreachable,
                },
                .op_amp => switch (c) {
                    '&' => {
                        return Token{ .tag = .op_and, .loc = loc };
                    },
                    else => unreachable,
                },
                .op_deref => switch (c) {
                    '>' => {
                        return Token{ .tag = .op_derefacc, .loc = loc };
                    },
                    else => unreachable,
                },
                .list_l => {
                    return Token{ .tag = .list_l, .loc = loc };
                },
                .list_r => {
                    return Token{ .tag = .list_r, .loc = loc };
                },
                .list_lz => switch (c) {
                    '(' => {
                        return Token{ .tag = .op_derefacc, .loc = loc };
                    },
                    else => unreachable,
                },
            };
        }
        return null;
    }
};

test "simple tokenization" {
    const example =
        \\t
        \\nil
        \\a = b
        \\"some string w/ escaped\""
    ;

    var tokz = Tokenizer.init(example);
    try std.testing.expectEqual(Token{ .loc = .{ .start = 0, .end = 1 }, .tag = .t }, tokz.next());
    try std.testing.expectEqual(Token{ .loc = .{ .start = 2, .end = 5 }, .tag = .nil }, tokz.next());
    try std.testing.expectEqual(Token{ .loc = .{ .start = 6, .end = 7 }, .tag = .sym }, tokz.next());
    try std.testing.expectEqual(Token{ .loc = .{ .start = 8, .end = 9 }, .tag = .assign }, tokz.next());
    try std.testing.expectEqual(Token{ .loc = .{ .start = 10, .end = 11 }, .tag = .sym }, tokz.next());
    try std.testing.expectEqual(Token{ .loc = .{ .start = 12, .end = 37 }, .tag = .str }, tokz.next());
}
